/* 
 *   Creation Date: <1998-11-20 16:18:20 samuel>
 *   Time-stamp: <2003/03/09 04:21:54 samuel>
 *   
 *	<context.c>
 *	
 *	MMU context allocation
 *   
 *   Copyright (C) 1998-2003 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */

#include "archinclude.h"
#include "alloc.h"
#include "mmu.h"
#include "mmu_contexts.h"
#include "misc.h"
#include "asmfuncs.h"
#include "emu.h"
#include "mtable.h"

int		_g_illegal_vsid;		/* accessible through get_illegal_sr() */ 
static int 	is_initialized;
static int 	s_next_mol_context;		/* linux_context << 4 + esid */

int
flush_all_PTEs( void )
{
	ulong ea, v, sdr1 = _get_sdr1();
	ulong *pte = phys_to_virt( sdr1 & ~0xffff );
	int npte = ((((sdr1 & 0x1ff) << 16) | 0xffff) + 1) / 8;
	int i, count=0;

	/* SDR1 might not be initialized (yet) on the 603 */
	if( !is_initialized || !sdr1 )
		return 0;

	for(i=0; i<npte; i++, pte+=2 ) {
		v = *pte;
		if( !(v & BIT(0)) )	/* test V-bit */
			continue;
		v = (v & ~BIT(0)) >> 7;
		v = (v - ((v & 0xf) * MUNGE_ESID_ADD)) * MUNGE_MUL_INVERSE;
		v = (v>>4) & CTX_MASK;

		if( v >= FIRST_MOL_CONTEXT && v <= LAST_MOL_CONTEXT ) {
 			*pte = 0;
			count++;
		}
	}

	/* Perform a tlbia */
	for( ea=0; ea <= (0x3f << 12); ea += 0x1000 )
		__tlbie(ea);
	
	if( count )
		printk("Flush_all_PTEs: %d\n", count );
	return count;
}

int
setup_mol_contexts( void )
{
	/* Remember: contexts live in the session-global namespace */
	if( is_initialized )
		return 0;

	s_next_mol_context = FIRST_MOL_CONTEXT << 4;
	/* XXX: We should consider setting the no-execute bit eventually.
	 * This ought to optimize away a hash table walk in certain cases.
	 */
	_g_illegal_vsid = munge_mol_context( s_next_mol_context++ ) | BIT(2) | BIT(3);

#ifndef NEW_LINUX_MM
	if( atomic_read( mc_next_mmu_context ) > LAST_LINUX_CONTEXT ) {
		printk("Context overflow\n");
		/* let the real overflow occur as soon as possible */
		atomic_set( mc_next_mmu_context, LAST_LINUX_CONTEXT );

		/* mostly unnecessary */
		flush_all_PTEs();
		is_initialized = 0;
		return 1;
	}
#endif
	is_initialized = 1;
	flush_all_PTEs();
	return 0;
}

void
release_mol_contexts( void )
{
	/* To be called when the last session exits! */
	flush_all_PTEs();
	is_initialized = 0;
}

int
alloc_mol_context( int n )
{
	int i, ret;

	if( s_next_mol_context + n > (LAST_MOL_CONTEXT << 4) ) {
		printk("MOL context overflow\n");

		for( i=0; i<MAX_NUM_SESSIONS; i++ ) {
			kernel_vars_t *kv;
			if( !(kv=g_sesstab->kvars[i]) )
				continue;
			clear_all_vsids( kv );
		}
		flush_all_PTEs();
		s_next_mol_context = FIRST_MOL_CONTEXT << 4;
	}
	ret = s_next_mol_context;
	s_next_mol_context += n;

	return ret;
}

int
munge_mol_context( int mol_context )
{
	int vsid = MUNGE_CONTEXT((mol_context >> 4)) << 4;
	vsid += MUNGE_ESID_ADD * (mol_context & 0xf);

	//printk("mol_vsid: %08X munged: %08X\n", mol_context, (vsid & VSID_MASK) );
	return (vsid & VSID_MASK);
}

#ifndef NEW_LINUX_MM
void
linux_context_overflow( void )
{
	int v = atomic_read( mc_next_mmu_context );
	
	if( v >= LAST_CONTEXT )
		return;
	if( v >= FIRST_MOL_CONTEXT )
		printk("MOL internal error: contexts overwritten\n");
	/* Linux will notice this and wrap around */
	atomic_set( mc_next_mmu_context, LAST_CONTEXT );
}
#endif
